from CBD import *

############
#
# Causal Block Diagram to DEVS Compiler
# This compiler takes a causal block diagram as input and produces a .py file representing a DEVS model of this CBD
#
#	@Author: Nicolas Demarbaix
#	@Date: 2014-2015
###########

class CBD2DevsCompiler:
	def __init__(self, cbd):
		"""
		initialize the compiler. 
		cbd - the CBD that will be compiled
		depgraph - the dependency graph corresponding to the CBD
		known_block_values - Contains the value for each block that can be determined before the simulation is run (by propagating the values)
		states - Contains the states inside the CBD atomic DEVS
		"""
		self.__cbd = cbd
		self.__depgraph = cbd.createDepGraphForCompile()
		self.__known_block_values = dict()

		self.__states = dict()

	def __allInfluencersKnown(self, block):
		"""
		Method to check whether all the influencers of a block are known
		i.e. whether the value of the influencers is present in self.__known_block_values
		"""
		for inf in self.__depgraph.getInfluencers(block):
			if type(inf) is list:
				if inf[0].getBlockName() not in self.__known_block_values:
					return False
			else:
				if inf.getBlockName() not in self.__known_block_values:
					return False

		return True

	def __getActionStatement(self, block):
		"""
		getActionStatement method returns the computation statement for a block.
		This statement will be used in the action_int_..._to_...() subfunction of intTransition()
		"""
		if block.getBlockType() == "ProductBlock":
			s = ""
			wait = True
			for inf in self.__depgraph.getInfluencers(block):
				if not wait:
					s += "* "
				s += "self.state.%s "%inf.getBlockName().translate(None, ".")
				wait = False
			return s
		elif block.getBlockType() == "AdderBlock":
			s = ""
			wait = True
			for inf in self.__depgraph.getInfluencers(block):
				if not wait:
					s += "+ "
				s += "self.state.%s "%inf.getBlockName().translate(None, ".")
				wait = False
			return s
		elif block.getBlockType() == "NegatorBlock":
			s = ""
			s += "-self.state.%s"%self.__depgraph.getInfluencers(block)[0].getBlockName().translate(None, ".")
			return s
		elif block.getBlockType() == "InverterBlock":
			s = "1/self.state.%s"%self.__depgraph.getInfluencers(block)[0].getBlockName().translate(None, ".")
			return s
		elif block.getBlockType() == "WireBlock":
			s = "self.state.%s"%self.__depgraph.getInfluencers(block)[0].getBlockName().translate(None, ".")
			return s
		elif block.getBlockType() == "TimeBlock":
			s = "self.state.%s + self.state.rate"%block.getBlockName().translate(None,".")
			return s
		elif block.getBlockType() == "DelayBlock":
			s = "self.state.%s"%block.getDependencies(1)[0].getBlockName().translate(None,".")
			return s
		else:
			return "None"

	def __getBlockValue(self, block):
		"""
		Returns the block value for a given block.
		Used for setting up the CBDState with the values known at initialization.
		"""
		if block.getBlockType() == "ConstantBlock":
			return block.getValue()
		elif block.getBlockType() == "WireBlock":
			inf_name = self.__depgraph.getInfluencers(block)[0].getBlockName()
			return self.__known_block_values[inf_name]
		elif block.getBlockType() == "NegatorBlock":
			inf_name = self.__depgraph.getInfluencers(block)[0].getBlockName()
			return (-(self.__known_block_values[inf_name]))
		elif block.getBlockType() == "InverterBlock":
			inf_name = self.__depgraph.getInfluencers(block)[0].getBlockName()
			val = self.__known_block_values[inf_name]
			if val == 0.0:
				return 0.0
			else:
				return 1/val
		elif block.getBlockType() == "AdderBlock":
			result = 0.0
			for inf in self.__depgraph.getInfluencers(block):
				result = result + self.__known_block_values[inf.getBlockName()]
			return result
		elif block.getBlockType() == "ProductBlock":
			result = 1.0
			for inf in self.__depgraph.getInfluencers(block):
				result = result * self.__known_block_values[inf.getBlockName()]
			return result 
		else:
			return 0.0

	def compileToDevs(self, filename="cbd_devs.py"):
		"""
		The main method that compiles a CBD to a DEVS
		"""
		if (".py" not in filename):
			filename = "%s%s" % filename, ".py"

		# Check if the CBD is hierarchical
		isHierarchical = False
		for block in self.__cbd.getBlocks():
			if block.getBlockType() == "CBD":
				isHierarchical = True
				break

		# if the CBD is hierarchical we first flatten the CBD
		if isHierarchical:
			self.__cbd.flatten()
			
		# Build the dependency graph
		self.__depgraph = self.__cbd.createDepGraphForCompile(1)


		f = open(filename, 'w')
		# Check if non linear component is found inside the CBD.
		# If a nonlinear component is present, this will actually result in the compilation ending, because CBD calls exit(1)
		# when a nonlinear loop is detected
		isNonLinear = False
		for component in self.__depgraph.getStrongComponents():
			if not self.__cbd.isLinear(component):
				isNonLinear = True
				break

		# Write the standard preface to the file
		f.write("""
from pypdevs.DEVS import *
from pypdevs.infinity import INFINITY
import random
import math
""")
		# The following dictionaries are provided to make sure the correct type checking, and dependency checking can be performed
		# in the constructLinearInput method.
		# The component dictionary is used to check whether a certain element belongs to a component (and thus should be calculated
		# when transitioning from a certain state to another)
		f.write("""
dependsDict = {""")
		wait = True
		for block in self.__cbd.getBlocks():
			if not wait:
				f.write(""", """)
			f.write("""\"%s\": ["""%block.getBlockName().translate(None,"."))
			if block.getBlockType() == "DelayBlock":
				f.write("""\"%s\"]"""%(block.getDependencies(1)[0].getBlockName().translate(None,".")))
			else:
				wait2 = True
				for dependsOn in self.__depgraph.getInfluencers(block):
					if not wait2:
						f.write(""", """)
					if type(dependsOn) is list:
						name = dependsOn[0].getBlockName()
						f.write("""\"%s\""""%name)
					else:
						f.write("""\"%s\""""%dependsOn.getBlockName().translate(None,"."))
					wait2 = False
				f.write("""]""")
				wait = False
		f.write("""}""")
		f.write("""
influencesDict = {""")
		wait = True
		for block in self.__cbd.getBlocks():
			if not wait:
				f.write(""", """)
			f.write("""\"%s\": ["""%block.getBlockName().translate(None,"."))
			wait2 = True
			for influenced in self.__depgraph.getDependents(block):
				if not wait2:
					f.write(""", """)
				if type(influenced) is list:
					f.write("""\"%s\""""%influenced[0].getBlockName().translate(None,"."))
				else:
					f.write("""\"%s\""""%influenced.getBlockName().translate(None,"."))
				wait2 = False
			f.write("""]""")
			wait = False
		f.write("""}""")

		f.write("""
componentDict = {""")
		wait = True
		index = 0
		for component in self.__depgraph.getStrongComponents():
			if not wait:
				f.write(""", """)
			f.write("""\"%d\": ["""%index)
			wait2 = True
			for block in component:
				if not wait2:
					f.write(""", """)
				f.write("""\"%s\""""%block.getBlockName().translate(None,"."))
				wait2 = False
			f.write("""]""")
			wait = False
			index += 1
		f.write("""}""")
		f.write("""
typeDict ={""")
		wait = True
		for block in self.__cbd.getBlocks():
			if not wait:
				f.write(""", """)
			f.write("""\"%s\": \"%s\""""%(block.getBlockName().translate(None,"."),block.getBlockType()))
			wait = False
		f.write("""}""")
		# Begin writing the CBDState functionality
		# Values of blocks that can be discovered during initialization are already filled in. Other blocks are referenced from the
		# dictionary that is passed from a transition action
		# An exception here is that for a TimeBlock and a DelayBlock these values are not directly filled in as they change dynamically
		f.write("""
class CBDState:
	def __init__(self, name=\"\", block_values=dict()):
		self.name = name
		self.rate = block_values[\"rate\"]
		self.iteration = block_values[\"iteration\"]""")

		for block in self.__cbd.getBlocks():
			if block.getBlockType() == "ConstantBlock":
				f.write("""
		self.%s = %1.4f"""%(block.getBlockName().translate(None,"."), block.getValue()))
				self.__known_block_values[block.getBlockName()] = block.getValue()

		for block in self.__cbd.getBlocks():
			if block.getBlockName() in self.__known_block_values:
				continue

			allInfKnown = self.__allInfluencersKnown(block)


			if allInfKnown and block.getBlockType() != "TimeBlock" and block.getBlockType() != "DelayBlock":
				val = self.__getBlockValue(block)
				f.write("""
		self.%s = %1.4f"""%(block.getBlockName().translate(None,"."), val))
				self.__known_block_values[block.getBlockName()] = val
			else:
				f.write("""
		self.%s = block_values[\"%s\"]"""%(block.getBlockName().translate(None,"."), block.getBlockName().translate(None,".")))

		f.write("""

	def __str__(self):
		s = \"\"
		s += \"state.name = \" + str(self.name) + \"\\n\"
		s += \"state.rate = \" + str(self.rate) + \"\\n\"
		s += \"state.iteration = \" + str(self.iteration) + \"\\n\"""")
		for block in self.__cbd.getBlocks():
			f.write("""
		s += \"state.%s = \" + str(self.%s) + \"\\n\""""%(block.getBlockName().translate(None,"."), block.getBlockName().translate(None,".")))
		f.write("""
		return s""")

		# The method get value is used in the constructLinearInput to retreive a value from the state
		# based on the block_name retreived from the dictionaries (see above).
		f.write("""
	def getValue(self, forBlock=""):""")
		for block in self.__cbd.getBlocks():
			f.write("""
		if forBlock == \"%s\":
			return self.%s"""%(block.getBlockName().translate(None,"."), block.getBlockName().translate(None,".")))
		f.write("""
		else:
			return 0.0""")

		prev = "start"
		index = 0
		for component in self.__depgraph.getStrongComponents():
			if self.__cbd.hasCycle(component, self.__depgraph):
				self.__states[prev] = "%d"%index
				prev = "%d"%index
			else:
				if component[0].getBlockName() not in self.__known_block_values and component[0].getBlockType() != "TimeBlock":
					self.__states[prev] = "%d"%index
					prev = "%d"%index
			index = index + 1
		self.__states[prev] = "start"

		# Begin writing the CBD class which is a single atomic devs containing states for all components that contain blocks that
		# still need to be calculated
		f.write("""
class CBD(AtomicDEVS):
	def __init__(self, name=\"CBD\"):
		AtomicDEVS.__init__(self,name)
		d = dict()""")
		for block in self.__cbd.getBlocks():
			if block.getBlockName() not in self.__known_block_values:
				hasDelayAsDependent = False
				for inf in self.__depgraph.getDependents(block):
					if inf.getBlockType() == "DelayBlock":
						hasDelayAsDependent = True
						break
				if hasDelayAsDependent:
					continue
				if block.getBlockType() == "TimeBlock":
					f.write("""
		d[\"%s\"] = 0.0"""%block.getBlockName().translate(None,"."))
				elif block.getBlockType() == "DelayBlock" and block.getDependencies(0)[0].getBlockName() in self.__known_block_values:
					f.write("""
		d[\"%s\"] = %1.4f"""%(block.getBlockName().translate(None,"."), self.__known_block_values[block.getDependencies(0)[0].getBlockName()]))
					f.write("""
		d[\"%s\"] = %1.4f"""%(block.getDependencies(1)[0].getBlockName().translate(None,"."), self.__known_block_values[block.getDependencies(0)[0].getBlockName()]))
				else:
					f.write("""
		d[\"%s\"] = None"""%block.getBlockName().translate(None,"."))
		f.write("""
		d[\"iteration\"] = 0
		d[\"rate\"] = 1.0""")
		f.write("""
		self.state = CBDState(name="start", block_values=d)
		self.my_ports = {}
			""")

		# Timeadvance method returns the same value for all states, namely the rate of the CBD
		# Normally this should be the rate of the block under consideration, but we assume this is general for the entire system
		f.write("""
	def timeAdvance(self):
		return self.state.rate""")

		# As there is only a single AtomicDEVS with no ouput ports, the output function returns an empty dictionary
		f.write("""

	def outputFunc(self):
		def subFunc(self):
			return {}
		return {self.my_ports[k]: v for k, v in subfunc(self).iteritems()}
			""")

		# Construct for each internal transition to corresponding condition and action
		f.write("""
	def intTransition(self):""")
		for key, value in self.__states.iteritems():
			if value == "start":
				f.write("""
		def cond_int_%s_to_%s():
			return True"""%(key, value))
			else:
				f.write("""
		def cond_int_%s_to_%s():
			if """%(key, value))
				comp = self.__depgraph.getStrongComponents()[int(value)]
				wait = True
				for block in comp:
					for inf in self.__depgraph.getInfluencers(block):
						if inf not in comp:
							if not wait:
								f.write("""and """)
							f.write("""self.state.%s != None """%inf.getBlockName().translate(None,"."))
							wait = False
				f.write(""":
				return True
			else:
				return False""")


			
			if value == "start":
				f.write("""
		def action_int_%s_to_%s():"""%(key, value))
				for block in self.__cbd.getBlocks():
					if block.getBlockType() == "ConstantBlock":
						continue
					elif block.getBlockType() == "TimeBlock":
						f.write("""
			val_%s = %s"""%(block.getBlockName().translate(None,"."), self.__getActionStatement(block)))
					else:
						f.write("""
			val_%s = self.state.%s"""%(block.getBlockName().translate(None,"."), block.getBlockName().translate(None,".")))

				f.write("""
			return {\"rate\": self.state.rate * self.updateRate(), \"iteration\": self.state.iteration + 1""")
				for block in self.__cbd.getBlocks():
					if block.getBlockName() not in self.__known_block_values:
						f.write(""", \"%s\": %s"""%(block.getBlockName().translate(None,"."), "val_%s"%block.getBlockName().translate(None,".")))
				f.write("""}""")
			else:
				f.write("""
		def action_int_%s_to_%s():"""%(key, value))
				comp = self.__depgraph.getStrongComponents()[int(value)]
				if self.__cbd.hasCycle(comp, self.__depgraph):
					f.write("""
			component = componentDict[\"%s\"]
			solverInput = self.constructLinearOutput(component)
			self.__gaussjLinearSolver(solverInput)
			solutionVector = solverInput[1]
			return {\"rate\": self.state.rate * self.updateRate(), \"iteration\": self.state.iteration"""%value)
					for block in self.__cbd.getBlocks():
						if (block.getBlockName() not in self.__known_block_values) and (block not in comp):
							f.write(""", \"%s\": self.state.%s"""%(block.getBlockName().translate(None,"."), block.getBlockName().translate(None,".")))
					index = 0
					for block in comp:
						f.write(""", \"%s\": solutionVector[%d]"""%(block.getBlockName().translate(None,"."), index))
						wait = False
						index += 1
				else:
					hasDelayAsDependent2 = False
					depend_name = ""
					if comp[0].getBlockType() == "TimeBlock":
						f.write("""
			val_%s = self.state.%s"""%(comp[0].getBlockName().translate(None,"."), comp[0].getBlockName().translate(None,".")))
					elif comp[0].getBlockType() == "DelayBlock":
						f.write("""
			val_%s = self.state.%s"""%(comp[0].getBlockName().translate(None,"."), comp[0].getBlockName().translate(None,".")))
					else:
						f.write("""
			val_%s = None
			if """%(comp[0].getBlockName().translate(None,".")))
						wait = True
						for inf in self.__depgraph.getInfluencers(comp[0]):
							if not wait:
								f.write(""" and """)
							f.write("""self.state.%s is not None"""%inf.getBlockName().translate(None,"."))
							wait = False
						f.write(""":
				val_%s = %s"""%(comp[0].getBlockName().translate(None,"."), self.__getActionStatement(comp[0])))
						dependents = self.__depgraph.getDependents(comp[0])
						if len(dependents) > 0:
							if dependents[0].getBlockType() == "DelayBlock":
								f.write("""
			val_%s = self.state.%s"""%(dependents[0].getBlockName().translate(None,"."), comp[0].getBlockName().translate(None,".")))
								hasDelayAsDependent2 = True
								depend_name = dependents[0].getBlockName()						

					f.write("""
			return {\"rate\": self.state.rate * self.updateRate(), \"iteration\": self.state.iteration""")
					f.write(""", \"%s\": %s"""%(comp[0].getBlockName().translate(None,"."), "val_%s"%comp[0].getBlockName().translate(None,".")))
					for block in self.__cbd.getBlocks():
						if hasDelayAsDependent2:
							if block.getBlockName() == depend_name:
								f.write(""",\"%s\": val_%s"""%(block.getBlockName().translate(None,"."), block.getBlockName().translate(None,".")))
								continue
						if (block.getBlockName() not in self.__known_block_values) and (block not in comp):
							f.write(""", \"%s\": self.state.%s """%(block.getBlockName().translate(None,"."), block.getBlockName().translate(None,".")))
				f.write("""}""")

			f.write("""
		if self.state.name == \"%s\" and cond_int_%s_to_%s():
			return CBDState(\"%s\", action_int_%s_to_%s())
				"""%(key, key, value, value, key, value))
		f.write("""
		else:
			return AtomicDEVS.intTransition(self)
			""")

		# No external transitions are used in this DEVS
		f.write("""
	def extTransition(self, my_inputs):
		return AtomicDEVS.extTransition(self, my_inputs)
			""")

		# The method updateRate can be used to dynamically change the rate of the CBD based on values in the CBD (i.e. in CBDState)
		f.write("""
	def updateRate(self):
		errorTooLarge = False

		# INSERT A CHECK HERE
		# e.g.
		#		if abs(self.state.derivOUT1 - 1.0) > 0.001: errorTooLarge = True

		# NOTE: Don't take the factor to small (e.g. 0.75) or the CBD will proceed very slow.
		changeFactor = 0.99
		if errorTooLarge:
			return changeFactor
		else:
			return 1.0
			""")

		# The constructLinearOutput method is used to compute the values of blocks that belong to a cyclic component.
		# It is based on the constructLinearOutput method found in the CBD class
		f.write("""
	def constructLinearOutput(self, strongComponent):
		size = len(strongComponent)
		row = []
		M1 = []
		M2 = []

		# INITIALIZE MATRICES
		i = 0
		while (i < size):
			j = 0
			row = []
			while (j < size):
				row.append(0)
				j += 1
			M1.append(row)
			M2.append(0)
			i += 1

		# block -> index of block
		indexdict = dict()

		for (i, block) in enumerate(strongComponent):
			indexdict[block] = i

		for (i, block) in enumerate(strongComponent):
			if typeDict[block] == \"AdderBlock\":
				for inf in [x for x in dependsDict[block] if x not in strongComponent]:
					M2[i] -= self.state.getValue(inf)
				M1[i][i] = -1
				for compInStrong in [x for x in dependsDict[block] if x in strongComponent]:
					M1[i][indexdict[compInStrong]] = 1
			elif typeDict[block] == \"ProductBlock\":
				M1[i][i] = -1
				M1[i][indexdict[[x for x in dependsDict[block] if x in strongComponent][0]]] = reduce(lambda x,y: x*y, [self.state.getValue(x) for x in dependsDict[block] if x not in strongComponent])
			elif typeDict[block] == \"NegatorBlock\":
				M1[i][i] = -1
				M1[i][indexdict[dependsDict[block][0]]] = -1
			elif typeDict[block] == \"InputPortBlock\":
				M1[i][i] = 1
				M1[i][indexdict[dependsDict[block][0]]] = -1
			elif typeDict[block] == \"OutputPortBlock\" or typeDict[block] == \"WireBlock\":
				M1[i][i] = 1
				M1[i][indexdict[dependsDict[block][0]]] = -1
			elif typeDict[block] == \"DelayBlock\":
				M2[i] -= self.state.getValue(dependsDict[block][0])
				M1[i][i] = -1
		return [M1, M2]
			""")
		
		# The following three methods (__ivector, __swap and __gaussjLinearSolver) are copied directly from the CBD class
		f.write("""
	def __ivector(self, n):
		v = []
		for i in range(n):
			v.append(0)
		return v

	def __swap(self, a, b):
		return (b, a)
			""")

		f.write("""
	def __gaussjLinearSolver(self, solverInput):
		M1 = solverInput[0]
		M2 = solverInput[1]
		n = len(M1)
		indxc = self.__ivector(n)
		indxr = self.__ivector(n)
		ipiv = self.__ivector(n)
		icol = 0
		irow = 0
		for i in range(n):
			big = 0.0
			for j in range(n):
				if (ipiv[j] != 1):
					for k in range(n):
						if (ipiv[k] == 0):
							if (math.fabs(M1[j][k]) >= big):
								big = math.fabs(M1[j][k])
								irow = j
								icol = k
						elif (ipiv[k] > 1):
							raise ValueError("GAUSSJ: Singular Matrix-1")
			ipiv[icol] += 1
			if (irow != icol):
				for l in range(n):
					(M1[irow][l], M1[icol][l]) = self.__swap(M1[irow][l], M1[icol][l])
				(M2[irow], M2[icol]) = self.__swap(M2[irow], M2[icol])
			indxr[i] = irow
			indxc[i] = icol
			if (M1[icol][icol] == 0.0):
				raise ValueError("GAUSSJ: Singular Matrix-2")
			pivinv = 1.0/M1[icol][icol]
			M1[icol][icol] = 1.0
			for l in range(n):
				M1[icol][l] *= pivinv
			M2[icol] *= pivinv
			for ll in range(n):
				if (ll != icol):
					dum = M1[ll][icol]
					M1[ll][icol] = 0.0
					for l in range(n):
						M1[ll][l] -= M1[icol][l] * dum
					M2[ll] -= M2[icol] * dum
		l = n
		while (l > 0):
			l -= 1
			if (indxr[l] != indxc[l]):
				for k in range(n):
					(M1[k][indxr[l]], M1[k][indxc[l]]) = self.__swap(M1[k][indxr[l]], M1[k][indxc[l]])
			""")

		# Begin writing the CoupledDEVS. It contains no ports and a single submodel, the CBD AtomicDEVS.
		f.write("""
class Root(CoupledDEVS):
	def __init__(self, name=\"Root\"):
		CoupledDEVS.__init__(self, name)

		self.my_ports = {}
		self.submodels = {}
		self.submodels[\"CBD\"] = self.addSubModel(CBD(name=\"CBD\"))
			""")

		f.write("""
	def select(self, imm):
		return imm[0]
			""")


		f.close()
		print "The CBD has been succesfully compiled to DEVS (%s)." % filename