from CBD import *
import re

############
#
# Causal Block Diagram to LaTeX Compiler
# This compiler takes a causal block diagram as input and produces a .tex file that contains the following information:
#	* Causal Block Diagram: 
#		- A textual description of the CBD (block names/types and the system of equations that describes the CBD)
#		- A graphical representation of the CBD using the tikz package for LaTeX
#	* Dependency Graph:
#		- A textual description of the dependency graph (The different components with the blocks that belong to these components,
#			for each block its dependencies and influencers)
#	* A solution for the CBD:
#		- If the CBD contains no loops: The output signals for each block are displayed (after one run)
#		- If the CBD contains linear loops: For each cyclic component the input matrices for the gaussian solver are provided along with the
#		  corresponding system of equations. The output signals are provided as well
#		- If the CBD contains nonlinear loops: A message is provided stating that CBD's with nonlinear loops cannot be processed.
#
#	@Author: Nicolas Demarbaix
#	@Date: 2014-2015
###########

class CBD2LatexCompiler:
	def __init__(self, cbd):
		"""
		Initialize the compiler with:
		cbd - the Causal Block Diagram
		depgraph - the dependency graph corresponding to the CBD
		tikzlevels - dictionary used for displaying the CBD using tikz
		"""
		self.__cbd = cbd
		self.__depgraph = cbd.createDepGraphForCompile()
		
		self.__tikzlevels = dict()

	def __setupTikz(self):
		"""
		The method setupTikz fills the dictionary tikzlevels.
		Each level (horizontal line of nodes in the image) contain at most 5 nodes. The levels are build based on dependencies.
		e.g. level 0 = all block without dependencies
			 level 1 = all block with dependencies to blocks in level 0
			 etc.
		"""
		index = 1
		blocks = []
		for block in self.__cbd.getBlocks():
			blocks.append(block.getBlockName())

		self.__tikzlevels["level0"] = []

		indeces = []
		i = 0
		index = 0
		for block in blocks:
			if (len(self.__tikzlevels["level%d"%index]) >= 5):
				index = index + 1
				self.__tikzlevels["level%d"%index] = []
			if len(self.__cbd.getBlockByName(block).getLinksIn()) == 0:
				self.__tikzlevels["level%d"%index].append(block)
				indeces.append(i)
			i = i + 1

		offset = 0
		for i in indeces:
			blocks.pop(i - offset)
			offset = offset + 1

		index = index + 1
		check = []
		for x in range(index):
			check = check + self.__tikzlevels["level%d"%x]

		while (len(blocks) != 0):
			prev_key = "level%d"%(index-1)
			key = "level%d"%(index)
			self.__tikzlevels[key] = []
			indeces = []
			ind = 0
			
			for block in blocks:
				if len(self.__tikzlevels[key]) >= 5:
					check = check + self.__tikzlevels[key]
					index = index + 1
					key = "level%d"%(index)
					self.__tikzlevels[key] = []

				for item in check:
					if self.__cbd.getBlockByName(block).blockIsConnectedToInput(self.__cbd.getBlockByName(item)):
						self.__tikzlevels[key].append(block)
						indeces.append(ind)
						break
				ind = ind +1

			offset = 0
			for i in indeces:
				blocks.pop(i - offset)
				offset = offset + 1

			index = index + 1
			check = check + self.__tikzlevels[key]

	def _getSignal(self, blockname, output_port = None):
		"""
		The method get signal returns the signal array of a block.
		This method is copied directly from the test cases in this project
		"""
		foundBlocks = [ block for block in self.__cbd.getBlocks() if block.getBlockName() == blockname ]
		numFoundBlocks = len(foundBlocks)
		if numFoundBlocks == 1:
			signal =  foundBlocks[0].getSignal(name_output = output_port)
			return [x.value for x in signal]
		else:
			raise Exception(str(numFoundBlocks) + " blocks with name " + blockname + " found.\nExpected a single block.")

	def compileToLatex(self, filename="cbd.tex"):
		"""
		The main method where the CBD is compiled to a .tex file format
		"""

		# Check if the CBD is hierarchical. If yes, the CBD will be flattened first
		isHierachical = False
		for block in self.__cbd.getBlocks():
			if block.getBlockType() == "CBD":
				isHierachical = True

		# Check whether the CBD contains any Integrator- or DerivatorBlocks. If yes, the output signals in 
		# section 3 of the .tex file will be displayed for 3 distinct iterations (they change over time)
		containsDerivOrIntegr = False
		for block in self.__cbd.getBlocks():
			if block.getBlockType() == "DerivatorBlock" or block.getBlockType() == "IntegratorBlock":
				containsDerivOrIntegr = True
				break

		if isHierachical:
			self.__cbd.flatten()

		# Create the depgraph.
		# We re-create the depgraph in case the CBD was flattened (new blocks are added to the CBD, so they
		# should be added to the depgraph as well)
		self.__depgraph = self.__cbd.createDepGraphForCompile()

		# Setup tikzlevels for the CBD
		self.__setupTikz()
		if (".tex" not in filename):
			filename = "%s%s" % filename, ".tex"

		# Begin with the standard preface of the document and the table containing blocks and their corresponding types
		f = open(filename, 'w')
		f.write("""
\documentclass{article}
\usepackage{mathtools}
\usepackage{tikz}
\usepackage{tikzscale}
\usepackage{filecontents}
\usetikzlibrary{arrows,automata,fit}
\\tikzset{my label/.style args={#1:#2}{
    append after command={
       (\\tikzlastnode.north) node [#1] {#2}
    }
  }
}
\\begin{document}
\\title{Generated LaTeX file for Causal Block Diagrams}

\section{Causal Block Diagram}
This first section gives a general description of the Causal Block Diagram in the form of a table as well as a visual representation.

\subsection{Description}
\\begin{center}
	\\begin{tabular}{| c | c |}
	\hline
	Name & Type \\\\
	\hline
		""")
		for block in self.__cbd.getBlocks():
			f.write("""
	%s & %s \\\\
	\hline"""%(block.getBlockName().translate(None, "\_"), block.getBlockType()))

		f.write("""
	\end{tabular}
\end{center}

The Causal Block Diagram is defined by the following set of equations:
$\\begin{dcases*}""")
		# For each block, write down the equation describing this block. (This equation is of the form "name = value" 
		# if the block is a ConstantBlock)
		for block in self.__cbd.getBlocks():
			opstring = ""
			skip = False
			if (block.getBlockType() == "ConstantBlock"):
				f.write("""
					%s = %1.4f\\\\"""%(block.getBlockName().translate(None,"\_"), block.getValue()))
				continue
			if (block.getBlockType() == "TimeBlock"):
				f.write("""
					%s = \"current time\"\\\\"""%(block.getBlockName().translate(None,"\_")))
				continue
			if (block.getBlockType() == "DelayBlock"):
				opstring = "; (after first iteration)"
			if (block.getBlockType() == "ProductBlock"):
				opstring = "*"
			if (block.getBlockType() == "AdderBlock"):
				opstring = "+"
			if (block.getBlockType() == "NegatorBlock"):
				f.write("""
	%s = """%(block.getBlockName()).translate(None, "\_"))
				for block_in in self.__cbd.getBlocks():
					if block.blockIsConnectedToInput(block_in):
						f.write("""-%s"""%(block_in.getBlockName().translate(None, "\_")))
						skip = True
						break
				
			if skip:
				f.write("""\\\\""")
				continue

			f.write("""
	%s = """%(block.getBlockName().translate(None, "\_")))
			wait = True
			for block_in in self.__cbd.getBlocks():
				if block.blockIsConnectedToInput(block_in):
					if wait:
						f.write("""%s"""%(block_in.getBlockName().translate(None, "\_")))
						wait = False
					else:
						f.write("""%s %s"""%(opstring, block_in.getBlockName().translate(None, "\_")))
			f.write("""\\\\""")
		f.write("""
\end{dcases*}$
\\newpage
\subsection{Graphical Representation}
See Figure 1 (might be at the end of the document due to its size).
\\begin{figure}[ht]
\centering
\\resizebox{0.9\\textwidth}{!}{%
\\begin{tikzpicture}[->,>=stealth',shorten >=1pt,auto,node distance=2.8cm,
                    semithick]
	\\tikzstyle{every state}=[fill=black!60!green,draw=none,text=black]
			""")
		first = True
		newFirst = False
		last = ""
		last_above = ""
		# Define the nodes in the tikzpicture per level
		for (x,y) in self.__tikzlevels.iteritems():
			for item in y:
				if first:
					f.write("""
	\\node[state] (%s)					{%s};"""%(item.translate(None, "\_").translate(None,"."),item.translate(None, "\_").translate(None,".")))
					last = item
					last_above = item
					first = False
				elif newFirst:
					f.write("""
	\\node[state] (%s) 	[below of=%s]	{%s};"""%(item.translate(None, "\_").translate(None,"."),last_above.translate(None, "\_").translate(None,"."),item.translate(None, "\_").translate(None,".")))
					last = item
					last_above = item
					newFirst = False
				else:
					f.write("""
	\\node[state] (%s)	[right of=%s]	{%s};"""%(item.translate(None, "\_").translate(None,"."), last.translate(None, "\_").translate(None,"."), item.translate(None, "\_").translate(None,".")))
					last = item
			newFirst = True


		# define the links between the nodes based on the links between blocks in the CBD
		for block in self.__cbd.getBlocks():
			f.write("""
	\path (%s)"""%(block.getBlockName().translate(None, "\_").translate(None,".")))
			for depend in self.__cbd.getBlocks():
				if depend.blockIsConnectedToInput(block):
					f.write("""
						edge [bend right] node {} (%s)
						"""%(depend.getBlockName().translate(None, "\_").translate(None,".")))
			f.write(""";""")

		f.write("""
\end{tikzpicture}
}%
\caption{Scheme of the Causal Block Diagram}
\end{figure}
			""")

#
# BEGIN WRITING DEPENDENCY GRAPH RELATED INFORMATION
#
		f.write("""
\\newpage
\section{Dependency Graph}
In this section a description and graphical representation of the dependency graph that corresponds to the Causal Block Diagram is provided.
\subsection{Description}
\\begin{center}
	\\begin{tabular}{| c | c |}
	\hline
	Component & Blocks \\\\
	\hline
	\hline""")
		# For the dependecy graph, construct the table containing components and their block
		index = 1
		for component in self.__depgraph.getStrongComponents():
			f.write("""
	%d & """%(index))
			wait = True
			for block in component:
				if not wait:
					f.write(""", """)
				f.write("""%s"""%(block.getBlockName().translate(None, "\_")))
				wait = False
			f.write("""\\\\
	\hline""")
			index = index + 1
		f.write("""	
	\end{tabular}
\end{center}
\\begin{center}
	\\begin{tabular}{| c | c | c |}
	\hline
	Block & Depends on & Influences\\\\
	\hline
	\hline""")
		# For the dependency graph, fill the table with entries (block, influencers, dependents)
		for component in self.__depgraph.getStrongComponents():
			for block in component:
				f.write("""
		%s & """%(block.getBlockName().translate(None, "\_")))
				wait = True
				for inf in self.__depgraph.getInfluencers(block):
					if not wait:
						f.write(""", """)
					if type(inf) is list:
						f.write("""%s"""%(inf[0].getBlockName().translate(None, "\_")))
					else:
						f.write("""%s"""%(inf.getBlockName().translate(None, "\_")))
					wait = False
				wait = True
				f.write(""" & """)
				for dep in self.__depgraph.getDependents(block):
					if not wait:
						f.write(""", """)
					f.write("""%s"""%(dep.getBlockName().translate(None, "\_")))
					wait = False
				f.write("""\\\\
		\hline""")
		f.write("""
	\end{tabular}
\end{center}
\\newpage
\subsection{Graphical Representation}
See Figure 2 (might be at the end of the document due to its size).
\\begin{figure}[htb]
\centering
\\resizebox{0.9\\textwidth}{!}{%
\\begin{tikzpicture}[->,>=stealth',shorten >=1pt,auto,node distance=2.8cm,
                    semithick]
	\\tikzstyle{every state}=[fill=black!60!green,draw=none,text=black]
			""")
		# Construct the tikz image for the dependecy graph.
		# Define the nodes in the image
		first = True
		newFirst = False
		last = ""
		last_above = ""
		for (x,y) in self.__tikzlevels.iteritems():
			for item in y:
				if first:
					f.write("""
	\\node[state] (%s)					{%s};"""%(item.translate(None,".").translate(None,"_"),item.translate(None,".").translate(None,"_")))
					last = item
					last_above = item
					first = False
				elif newFirst:
					f.write("""
	\\node[state] (%s) 	[below of=%s]	{%s};"""%(item.translate(None,".").translate(None,"_"),last_above.translate(None,".").translate(None,"_"),item.translate(None,".").translate(None,"_")))
					last = item
					last_above = item
					newFirst = False
				else:
					f.write("""
	\\node[state] (%s)	[right of=%s]	{%s};"""%(item.translate(None,".").translate(None,"_"), last.translate(None,".").translate(None,"_"), item.translate(None,".").translate(None,"_")))
					last = item
			newFirst = True

		# Find the dependencies between nodes in the dependency graph and define the links in the tikz image
		dependencies = dict()
		for component in self.__depgraph.getStrongComponents():
			for block in component:
				dependencies[block.getBlockName()] = []

		for component in self.__depgraph.getStrongComponents():
			for block in component:
				for dep in self.__depgraph.getDependents(block):
					dependencies[dep.getBlockName()].append(block.getBlockName())

		for (x, y) in dependencies.iteritems():
			f.write("""
	\path (%s)"""%(x.translate(None,".").translate(None,"_")))
			for item in y:
				f.write("""
						edge [bend right] node {} (%s)
						"""%(item.translate(None,".").translate(None,"_")))
			f.write(""";""")
		f.write("""
\end{tikzpicture}
}%
\caption{The depency graph}
\end{figure}
\\newpage
\section{Solution of the Causal Block Diagram}
			""")
		# To find a solution for the CBD, first check whether it contains a loop
		containsLoop = False
		for component in self.__depgraph.getStrongComponents():
			if (self.__cbd.hasCycle(component, self.__depgraph)):
				containsLoop = True
				break

		# If the CBD does not contain a loop, print the standard message along with the output signal values for each block.
		if not containsLoop:
			f.write("""
The Causal Block Diagram does not contain any algebraic loops. Therefor, a solution can be found by simply propagating values according to the system of equations found in the first section. This gives the following result.
				""")
			times = 1

			if containsDerivOrIntegr:
				times = 3
				f.write("""
\\noindent
\\textit{Note}:
The Causal Block Diagram did contain a Derivator- or an IntegratorBlock. Therefore, the simulation will be run for 3 iterations, and the result for each block will be displayed per iteration (each table representing an iteration).
					""")
			# run the cbd simulator to collect block signal values
			self.__cbd.run(times)
			f.write("""
\\begin{center}
\\begin{tabular} {| c |""")
			for x in range(times):
 				f.write(""" c |""")
 			f.write("""}
\hline
Block""")
			for x in range(times):
				f.write("""& Iteration %d """%(x))
			f.write("""\\\\
\hline
\hline""")
			for component in self.__depgraph.getStrongComponents():
				for block in component:
					f.write("""
	%s """%(block.getBlockName().translate(None, "\_")))
					for x in range(times):
						f.write("""& %1.5f"""%(self._getSignal(block.getBlockName())[x]))
					f.write("""\\\\
\hline""")
			f.write("""
\hline
\end{tabular}
\end{center}
				""")

		else:
			# Verify whether the loops in the CBD are linear or not
			isNonLinear = False
			for component in self.__depgraph.getStrongComponents():
				if not self.__cbd.isLinear(component):
					isNonLinear = True
					break
			# If not linear loop detected: Write the standard message displaying this will not work
			if isNonLinear:
				f.write("""
The Causal Block Diagram contains a \\textbf{non linear} algebraic loop. A solution cannot be found using the Gaussian Solver.
					""")
			else:
				f.write("""
The algebraic loop(s) in the Causal Block Diagram is (are) all linear. Therefore a solution can be found using the Gaussian Solver. The following matrix will be used as input value for the solver:
					""")

				times = 1
				if containsDerivOrIntegr:
					times = 3
					f.write("""
\\noindent
\\textit{Note}:
The Causal Block Diagram did contain a Derivator- or an IntegratorBlock. Therefore, the simulation will be run for 3 iterations, and the result for each block will be displayed per iteration (each table representing an iteration).
					""")
				# Run for compile works in exactly the same manner as the default run method of CBD. 
				# In this case, the solver matrices are returned such that they can be displayed properly
				solver = self.__cbd.runForCompile(3)

				index = 1
				for k in range(len(self.__depgraph.getStrongComponents())):
					if ("%d"%index) in solver:
						f.write(""" 
\\textbf{Gaussian Solver input for component %d}
							"""%(index))
						mA = solver["%d"%index][0]
						mB = solver["%d"%index][1]
						mSize = len(mA)

						f.write("""
$A_{%d, %d} = 
\\begin{bmatrix}
							"""%(mSize, mSize))
						wait = True
						for i in range(mSize):
							for j in range(mSize):
								if not wait:
									f.write(""" & %1.3f"""%(mA[i][j]))
								else:
									f.write("""%1.3f"""%(mA[i][j]))
									wait = False
							wait = True
							f.write("""\\\\
								""")
						f.write("""
\end{bmatrix}$  """)
						f.write("""
$B_{1, %d} = 
\\begin{bmatrix}
							"""%(mSize))
						for i in range(mSize):
							f.write("""%1.3f\\\\
								"""%(mB[i]))
						f.write("""
\end{bmatrix}$
\\\\
\\\\
							""")

						f.write("""
These input matrices correspond to the following system of equations:
\\\\
$\\begin{dcases*}""")
						component = self.__depgraph.getStrongComponents()[index-1]
						wait = True
						for i in range(mSize):
							for j in range(mSize):
								if not wait: 
									f.write(""" + (%1.3f*%s)"""%(mA[i][j], component[j].getBlockName()))
								else:
									f.write("""
(%1.3f*%s)"""%(mA[i][j], component[j].getBlockName()))
									wait = False
							wait = True
							f.write(""" = %1.3f\\\\"""%mB[i])
						f.write("""
\end{dcases*}$
\\\\
\\\\
							""")

					index = index + 1
				f.write("""
The resulting values for the different blocks in the Causal Block Diagram are the following
\\begin{center}
\\begin{tabular} {| c |""")
				for x in range(times):
					f.write(""" c |""")
				f.write("""}
\hline
Block""")
				for x in range(times):
					f.write(""" & Iteration %d """%x)
				f.write("""\\\\
\hline
\hline""")
				for component in self.__depgraph.getStrongComponents():
					for block in component:
						f.write("""
	%s """%(block.getBlockName().translate(None, "\_")))
						for x in range(times):
							f.write("""& %1.4f"""%(self._getSignal(block.getBlockName())[x]))
						f.write("""\\\\
\hline""")
				f.write("""
\end{tabular}
\end{center}
				""")

		f.write("""
\end{document}
			""")
		f.close()

		print "The CBD has been succesfully compiled to LaTeX (%s)." % filename